# -*- coding: utf-8 -*-#
# Python script for finding and checking the Cytometer output files. It will create a Consistency check excel file.

# @author: Sjoerd Visser, APE GmbH, Berlin
# @author: Konrad von Volkmann, APE GmbH, Berlin
# @author: Toralf Kaiser, DRFZ, Berlin

# -------------------------------------
# Don't change or delete this section
# -------------------------------------
# @inistart

# [script]
# name = 2. Data Consistency

# [param1]
# name = Directory
# label = Evaluation Directory
# value = C:\Temp\Messungen
# global_var = DataDir
# mandatory = true
# datatype = directory

# @iniend

# -------------------------------------
# Preparations
# -------------------------------------

import fcsparser
import numpy as np
import pandas as pd
import argparse
import configparser
import glob, os, fnmatch, re

from ape.gui import Control

# Excel packages?
from openpyxl import Workbook
from openpyxl import load_workbook
from openpyxl import formatting, styles
from openpyxl.styles import Alignment
from openpyxl.utils import get_column_letter


import warnings
warnings.filterwarnings("ignore")

# ----------------------
# Functions
# ----------------------

def build_df_laser(meta):
    next_laser = True
    i = 1
    laser_name = []
    laser_delay = []
    laser_AreaScalingFactor = []
    while next_laser:
        try:
            key_str = 'LASER%dNAME' % i
            laser_name.append(meta[key_str])
            key_str = 'LASER%dDELAY' % i
            laser_delay.append(float(meta[key_str]))
            key_str = 'LASER%dASF' % i
            laser_AreaScalingFactor.append(float(meta[key_str]))
            i += 1
        except:
            next_laser = False
    df_laser = pd.DataFrame({'laser name':laser_name,'laser delay':laser_delay,'area scaling factor':laser_AreaScalingFactor})
    return df_laser

def style_range(ws, cell_range, border=styles.Border(), fill=None, font=None, alignment=None):
    """
    Apply styles to a range of cells as if they were a single cell.

    :param ws:  Excel worksheet instance
    :param range: An excel range to style (e.g. A1:F20)
    :param border: An openpyxl Border
    :param fill: An openpyxl PatternFill or GradientFill
    :param font: An openpyxl Font object
    """

    top = styles.Border(top=border.top)
    left =styles.Border(left=border.left)
    right = styles.Border(right=border.right)
    bottom = styles.Border(bottom=border.bottom)

    first_cell = ws[cell_range.split(":")[0]]
    if alignment:
        ws.merge_cells(cell_range)
        first_cell.alignment = alignment

    rows = ws[cell_range]
    if font:
        first_cell.font = font

    for cell in rows[0]:
        cell.border = cell.border + top
    for cell in rows[-1]:
        cell.border = cell.border + bottom

    for row in rows:
        l = row[0]
        r = row[-1]
        l.border = l.border + left
        r.border = r.border + right
        if fill:
            for c in row:
                c.fill = fill

def style_cell(ws, cell_range, border=styles.Border(), fill=None, font=None, alignment=None):
    """
    Apply styles to a range of cells as if they were a single cell.

    :param ws:  Excel worksheet instance
    :param range: An excel range to style (e.g. A1:F20)
    :param border: An openpyxl Border
    :param fill: An openpyxl PatternFill or GradientFill
    :param font: An openpyxl Font object
    """

    top = styles.Border(top=border.top)
    left =styles.Border(left=border.left)
    right = styles.Border(right=border.right)
    bottom = styles.Border(bottom=border.bottom)

    cell = ws[cell_range]

    if font:
        cell.font = font

        cell.border = cell.border + top
        cell.border = cell.border + bottom
        cell.border = cell.border + left
        cell.border = cell.border + right
        if fill:
            cell.fill = fill

def Excel_title(ws, Cells, Title):
    """
    Making the Banner on a certain Excel page

    :param ws:  Excel worksheet instance
    :param Title: str with title
    """

    blue_fill = styles.PatternFill("solid", fgColor="4398ED")
    thin = styles.Side(border_style="thin", color="000000")
    border1 = styles.Border(top=thin, left=thin, right=thin, bottom=thin)


    #Text
    White_color_font = 'FFFFFF'
    White_font = styles.Font(size=28, bold=True, color=White_color_font)


    style_range(ws, Cells, border=border1, fill=blue_fill, font=White_font, alignment=None)
    ws.merge_cells(Cells)
    text_placement = Cells.split(':')[0]
    ws[text_placement] = Title
    ws[text_placement].alignment = Alignment(horizontal='left',wrapText=True)
    ws.row_dimensions[1].height = 40

def Excel_Infobox(ws, Cells, text):
    """
    Making an information box on a certain Excel page

    :param ws:  Excel worksheet instance
    :param text: str with text
    """

    white_fill = styles.PatternFill("solid", fgColor="FFFFFF")
    thin = styles.Side(border_style="thin", color="000000")
    border1 = styles.Border(top=thin, left=thin, right=thin, bottom=thin)


    #Text
    Black_color_font = '000000'
    Black_font = styles.Font(size=12, bold=False, color=Black_color_font)


    style_range(ws, Cells, border=border1, fill=white_fill, font=Black_font, alignment=None)
    ws.merge_cells(Cells)
    text_placement = Cells.split(':')[0]
    ws[text_placement] = text
    ws[text_placement].alignment = Alignment(horizontal='left',wrapText=True)

# ======================
# Intensity ramp
# ======================

argparser = argparse.ArgumentParser(description='Description')
argparser.add_argument('--Directory', metavar='Directory', type=str)
args = argparser.parse_args()
data_folder = args.Directory

gui = Control()

# ----------------------
# Start script
# ----------------------

print('------------------------------------')
print('Script: 2: Create Consisteny file')

print('------------------------------------')
print('\n')


# ----------------------
# Load Data
# ----------------------
# analyze "ramp" folder
ramp_folder = os.path.join(data_folder, 'ramp')
ramp_filter = os.path.join(ramp_folder, '*.fcs')
lib_folder = os.path.join(data_folder,'LIB')
lib_filter = os.path.join(lib_folder,'*.fcs')
cfg_files = os.path.join(data_folder,'quantiFlash_script_settings*.cfg')


# ----------------------
# Check if folders and data structure
# ----------------------
if not os.path.exists(data_folder):
    gui.error('Data folder not found.\nPlease check the "Evaluation Directory"')
else:
    pass
if not os.path.exists(ramp_folder):
    gui.error('Ramp data not found.\nPlease check the correct data structure')
else:
    ramp_files = fnmatch.filter(os.listdir(ramp_folder), '*.fcs')
    if np.size(ramp_files) == 0:
        gui.error('Ramp data not found.\nPlease check the correct data structure')
    else:
        print('Ramp files (fcs): %d' % np.size(ramp_files))
        
if not os.path.exists(lib_folder):
    gui.error('LIB data not found.\nPlease check the correct data structure')
else:
    lib_files = fnmatch.filter(os.listdir(lib_folder), '*.fcs')
    if np.size(lib_files) == 0:
        gui.error('LIB data not found.\nPlease check the correct data structure')
    else:
        print('Lib files (fcs): %d' % np.size(lib_files))

# Check for Evaluation folder, if not -> create
directory = os.path.join(data_folder,'Evaluation')
# if not os.path.exists(directory):
    # gui.error('Evaluation folder not found.\nPlease run the script "2. Data Consistency" first')

# ----------------------
# Load qF settings file (cfg) from main data folder
# ----------------------
cfgFilter = os.path.join(data_folder,'quantiFlash_script_settings*.cfg')
cfgFiles = glob.glob(cfgFilter)
# cfgFile = os.path.join(data_folder,'quantiFlash_script_settings.cfg')
if os.path.isfile(cfgFiles[0]):
    Config = configparser.ConfigParser()
    Config.read(cfgFiles[0])

    # Device settings
    device = str(Config.get('Main','device'))
    port_id = int(Config.get('Main','port'))

    # QuantiFlash settings
    startInt = float(Config.get('quantiFlash','startInt'))
    stopInt = float(Config.get('quantiFlash','stopInt'))
    stepInt = float(Config.get('quantiFlash','stepInt'))
    triggerInt = float(Config.get('quantiFlash','triggerInt'))
    countRate = float(Config.get('quantiFlash','countRate'))
    nSamples = float(Config.get('quantiFlash','nSamples'))
    time_to_sleep = float(nSamples/countRate)-0.1

    # Evaluation settings
    detector_select = str(Config.get('Evaluation','detectors'))
    resolution = int(Config.get('Evaluation','resolution'))
    V_ref = int(Config.get('Evaluation','v_ref'))

    # Output settings
    raw_SD = bool(Config.get('Output','raw_SD'))
    raw_mean = bool(Config.get('Output','raw_mean'))
    raw_median = bool(Config.get('Output','raw_median'))
    raw_LIB = True
    K_factor = bool(Config.get('Output','K_factor'))
    result = bool(Config.get('Output','result'))
else:
    if gui.confirm("No quantiFlash settings file found.\n Continue with default parameters?") != 'OK':
        sys.exit(1)
    else:
        startInt = -94
        stopInt = 0
        stepInt = 2
        triggerInt = 50
        countRate = 4000.0
        nSamples = 10000.0
        time_to_sleep = float(nSamples/countRate)

        # Evaluation settings
        detector_select = 'height'
        resolution = 262144
        V_ref = 550

        # Output settings
        raw_SD = True
        raw_mean = True
        raw_median = False
        raw_LIB = True
        K_factor = True
        result = True

# Ramp_count = 0
# Lib_count = 0
# for file in glob.glob(ramp_filter):
    # if file[-3:]=='fcs':
        # Ramp_count +=1
# for file in glob.glob(lib_filter):
    # if file[-3:]=='fcs':
        # Lib_count +=1

# print('Ramp files (fcs): ' + str(Ramp_count))
# print('Lib files (fcs): ' + str(Lib_count))
# print('\n')

# ramp_files_list=[]
# lib_files_list=[]


print('\n')
print('------------------------------------')
print('\n')

counter = 0
NN = np.size(ramp_files) + np.size(lib_files)
first_data_set = True
ctr = 0
ch_volt_from_name = []

for file_name in ramp_files:
    # folder, file_name = os.path.split(file)
    # print(file_name)
    # ramp_files_list.append(file_name)
    # Retreive Meta data and event data from fcs files
    
    test = re.findall(r'V\d+',file_name)
    if len(test)!=0:
        voltage_from_filename = int(test[0][1:])
    else:
        test = re.findall(r'R\d+',file_name)
        if len(test)!=0:
            voltage_from_filename = int(test[0][1:])
        else:
            voltage_from_filename = np.nan
    ch_volt_from_name.append(voltage_from_filename)        

    file = os.path.join(ramp_folder,file_name)
    meta, df_data = fcsparser.parse(file, reformat_meta=True)

    # find installed lasers
    next_laser = True
    i = 1
    laser_name = []
    laser_delay = []
    laser_AreaScalingFactor = []
    while next_laser:
        try:
            key_str = 'LASER%dNAME' % i
            laser_name.append(meta[key_str])
            i += 1
        except:
            next_laser = False

    for lasercounter in np.arange(1,i):
        try:
            key_str = 'LASER%dDELAY' % lasercounter
            laser_delay.append(float(meta[key_str]))
        except:
            laser_delay.append(None)
            print('no laser delay factor found')
        try:
            key_str = 'LASER%dASF' % lasercounter
            laser_AreaScalingFactor.append(float(meta[key_str]))
        except:
            laser_AreaScalingFactor.append(None)
            print('no area scaling factor found')


    # Data frames containing the information
    df_laser = pd.DataFrame({'laser name': laser_name, 'laser delay': laser_delay, 'area scaling factor': laser_AreaScalingFactor})

    # find out what kind of cytometer was used
    try:
        Device=meta['$CYT']
    except:
        Device='None'


    # find active channels (without -A or -H suffix)
    channels = meta['_channels_']
    Time = []
    Diff = []
    Int_step =[]
    Samples = []


    df_data['Diff'] = df_data['Time']-df_data['Time'].shift(1)
    df_data['Diff'].loc[0:10]=0

    # Seperate the intensities via a jump in time diff > 5
    Samples = df_data[df_data.Diff>5]
    Samples = Samples.index.values
    Samples = np.insert(Samples, 0,0)
    Samples = np.append(Samples, df_data.index[-1])
    Samples = np.diff(Samples)

    # First step is always off
    Int_step=['Off']
    for i in np.arange(startInt,stopInt+stepInt,stepInt):
        Int_step.append(str(i))

    df_temp = pd.DataFrame(Samples)
    temp2 = np.sum(Samples)

    if first_data_set:
        df_Int = df_temp
        ramp_total = temp2
    else:
        df_Int = pd.concat([df_Int,df_temp],axis=1)
        ramp_total = np.append(ramp_total, temp2)


    channels = channels[~channels['$PnN'].isin(np.array(['Time']))]
    channel_names_ar = channels['$PnN'].unique()
    for i in np.arange(channel_names_ar.shape[0]):
        channel_names_ar[i] = channel_names_ar[i][:-2]
    df_tmp = pd.DataFrame({'name': channel_names_ar})
    channel_names_ar = df_tmp['name'].unique()

    # initialize df_channels
    ch_volt = []
    ch_names = []
    # ch_volt_from_name = []
    ch_volt_compare = []

    # Check which channels are being recorded
    N = channel_names_ar.shape[0]
    ch_res = np.zeros(N)
    ch_area = np.zeros(N,'bool')
    ch_height = np.zeros(N,'bool')
    ch_PMT = np.zeros(N,'bool')
    df_channels = pd.DataFrame({'Resolution':ch_res,'Area':ch_area,'Height':ch_height,'PMT':ch_PMT},index=channel_names_ar)
    # build df_channels
    ch_idx = channels.index
    for i in ch_idx:
        check_volt=False
        ch_name = channels.loc[i,'$PnN'][:-2]
        ch_type = channels.loc[i,'$PnN'][-1]
        if ch_type == 'A':
            # Add Area to Channels data frame
            df_channels.loc[ch_name,'Area'] = True
            check_volt=True  #check this detector for voltage info
        elif ch_type == 'H':
            # Add Height to Channels data frame
            df_channels.loc[ch_name,'Height'] = True
            check_volt=True  #check this detector for voltage info
        if ch_name == 'FSC':
            # Add Front Scattering channel to Channels data frame
            df_channels.loc[ch_name,'PMT'] = False
            check_volt=False  #check this detector for voltage info
        else:
            # Add PMT to Channels data frame
            df_channels.loc[ch_name,'PMT'] = True
            check_volt=True  #check this detector for voltage info

        # Add Resolution to Channels data frame
        df_channels.loc[ch_name,'Resolution'] = channels.loc[i,'$PnR']

        if check_volt:  #check this detector for voltage info
            try:
                ch_volt.append(int(channels.loc[i,'$PnV']))
            except:
                try:
                    ch_volt.append(int(meta['$P%dV' %i]))
                except:
                    ch_volt.append(0)
        if check_volt:  #check this detector for voltage info
            try:
                ch_volt_compare.append(int(channels.loc[i,'$PnV'])==voltage_from_filename)
            except:
                try:
                    ch_volt_compare.append(int(meta['$P%dV' %i])==voltage_from_filename)
                except:
                    ch_volt_compare.append(False)

        if ch_name in ch_names:
            continue
        else:
            ch_names.append(ch_name)

        #new channel found, try to find voltage information


    channels = channels[~channels['$PnN'].isin(np.array(['FSC-A','FSC-H','Time']))]
    # Create Data frame for
    df_tmp = pd.DataFrame({file_name:ch_volt},index=channels['$PnN'])
    if first_data_set:
        df_channels_volt = df_tmp
    else:
        df_channels_volt = pd.concat([df_channels_volt,df_tmp],axis=1)

    df_tmp = pd.DataFrame({voltage_from_filename:ch_volt_compare},index=channels['$PnN'])
    if first_data_set:
        # df_tmp = pd.DataFrame({voltage_from_filename:ch_volt_compare,'channels':channels['$PnN']})
        df_volt_compare = df_tmp
    else:
        # df_tmp = pd.DataFrame({voltage_from_filename:ch_volt_compare})
        df_volt_compare = pd.concat([df_volt_compare,df_tmp],axis=1)

    PMT_error=False
    PMT_error_str = ""
    PMT_error_handling_str = ""
    most_common_V = max(set(ch_volt), key=ch_volt.count)

    # check if all channels are at set-voltage
    for i in np.arange(0,np.size(channels,0)):
        if ch_volt[i]!=most_common_V: # if channel voltage is not set voltage
            PMT_error =True
            PMT_error_str = PMT_error_str + channels.loc[3+i,'$PnN'] + ', '

    if PMT_error:
        PMT_error_handling_str = 'Channel(s) will be ignored'
    else:
        PMT_error_str = 'OK'

    # check if the laser delay is set to zero for all channels
    laser_delay_error = df_laser[df_laser['laser delay']!=0]
    if laser_delay_error.empty:
        Laser_delay_error_str = 'OK'
    else:
        Laser_delay_error_str = 'Error: '
        for tmp in laser_delay_error['laser name']:
            Laser_delay_error_str = Laser_delay_error_str + tmp + ', '

    # build DataFrame
    if first_data_set:
        try:
            Device=meta['$CYT']
        except:
            Device='None'
        try:
            DeviceID=meta['CYTNUM']
        except:
            DeviceID='None'
        df_meta = pd.DataFrame({'File':[file_name],'Device':[Device],'Device ID':[DeviceID],'Date':[meta['$DATE']],'Set voltage':[str(most_common_V)],'PMT voltage':[PMT_error_str],'Error handling':[PMT_error_handling_str],'Laser delay':Laser_delay_error_str})
    else:
        df_tmp = pd.DataFrame({'File':[file_name],'Device':[Device],'Device ID':[DeviceID],'Date':[meta['$DATE']],'Set voltage':[str(most_common_V)],'PMT voltage':[PMT_error_str],'Error handling':[PMT_error_handling_str],'Laser delay':Laser_delay_error_str})
        df_meta = pd.concat([df_meta,df_tmp],axis=0)
    first_data_set = False
    ctr += 1
    col_names = df_data.columns
    gui.progress("{:f}".format(counter/NN*100))
    counter += 1

tmp = df_Int.isna().sum()==0
tmp = np.array(tmp)
df_file_count = pd.DataFrame({'ramp_files':ramp_files,'ramp_complete':tmp},index=ch_volt_from_name)
df_volt_compare = df_volt_compare.T
df_data_save = df_data.copy()
Samples_save = Samples
df_Int_save = df_Int.copy()
ramp_total_save = ramp_total.copy()

# ----------------------
# Section to evatuate Lib files
# ----------------------

# analyze "LIB" folder
# first_data_set = False
ctr = 0
lib_total =[]
ch_volt_from_name = []
for file_name in lib_files:
    # folder, file_name = os.path.split(file)
    # print(file_name)
    # lib_files_list.append(file_name)
    
    test = re.findall(r'V\d+',file_name)
    if len(test)!=0:
        voltage_from_filename = int(test[0][1:])
    else:
        test = re.findall(r'B\d+',file_name)
        if len(test)!=0:
            voltage_from_filename = int(test[0][1:])
        else:
            voltage_from_filename = np.nan
    ch_volt_from_name.append(voltage_from_filename)        

    file = os.path.join(lib_folder,file_name)
    meta, df_data = fcsparser.parse(file, reformat_meta=True)

    # find installed lasers
    next_laser = True
    i = 1
    laser_name = []
    laser_delay = []
    laser_AreaScalingFactor = []
    while next_laser:
        try:
            key_str = 'LASER%dNAME' % i
            laser_name.append(meta[key_str])
            i += 1
        except:
            next_laser = False

    for lasercounter in np.arange(1,i):
        try:
            key_str = 'LASER%dDELAY' % lasercounter
            laser_delay.append(float(meta[key_str]))
        except:
            laser_delay.append(None)
            print('no laser delay factor found')
        try:
            key_str = 'LASER%dASF' % lasercounter
            laser_AreaScalingFactor.append(float(meta[key_str]))
        except:
            laser_AreaScalingFactor.append(None)
            print('no area scaling factor found')


    #data frames containing the information
    df_laser = pd.DataFrame({'laser name': laser_name, 'laser delay': laser_delay, 'area scaling factor': laser_AreaScalingFactor})

    temp = df_data['FSC-A'].size


    lib_total = np.append(lib_total, temp)


    # find active channels (without -A or -H suffix)
    channels = meta['_channels_']
    channels = channels[~channels['$PnN'].isin(np.array(['Time']))]
    channel_names_ar = channels['$PnN'].unique()
    for i in np.arange(channel_names_ar.shape[0]):
        channel_names_ar[i] = channel_names_ar[i][:-2]
    df_tmp = pd.DataFrame({'name': channel_names_ar})
    channel_names_ar = df_tmp['name'].unique()

    # initialize df_channels
    ch_volt = []
    ch_names = []
    # initialize df_channels
    N = channel_names_ar.shape[0]
    ch_res = np.zeros(N)
    ch_area = np.zeros(N,'bool')
    ch_height = np.zeros(N,'bool')
    ch_PMT = np.zeros(N,'bool')
    df_channels = pd.DataFrame({'Resolution':ch_res,'Area':ch_area,'Height':ch_height,'PMT':ch_PMT},index=channel_names_ar)
    # build df_channels
    ch_idx = channels.index
    for i in ch_idx:
        ch_name = channels.loc[i,'$PnN'][:-2]
        ch_type = channels.loc[i,'$PnN'][-1]


        check_volt=False
        if ch_type == 'A':
            # Add Area to Channels data frame
            df_channels.loc[ch_name,'Area'] = 1
            check_volt=True  #check this detector for voltage info
        elif ch_type == 'H':
            # Add Height to Channels data frame
            df_channels.loc[ch_name,'Height'] = 1
            check_volt=True  #check this detector for voltage info
        if ch_name == 'FSC':
            # Add Front Scattering channel to Channels data frame
            df_channels.loc[ch_name,'PMT'] = 0
            check_volt=False  #check this detector for voltage info
        else:
            # Add PMT to Channels data frame
            df_channels.loc[ch_name,'PMT'] = 1
            check_volt=True  #check this detector for voltage info

        # Add Resolution to Channels data frame
        df_channels.loc[ch_name,'Resolution'] = channels.loc[i,'$PnR']

        if check_volt:  #check this detector for voltage info
            try:
                ch_volt.append(int(channels.loc[i,'$PnV']))
            except:
                try:
                    ch_volt.append(int(meta['$P%dV' %i]))
                except:
                    ch_volt.append(0)


        if ch_name in ch_names:
            continue
        else:
            ch_names.append(ch_name)

        #new channel found, try to find voltage information


    channels = channels[~channels['$PnN'].isin(np.array(['FSC-A','FSC-H','Time']))]
    # Create Data frame for
    df_tmp = pd.DataFrame({file_name:ch_volt},index=channels['$PnN'])
    if first_data_set:
        df_channels_volt = df_tmp
    else:
        df_channels_volt = pd.concat([df_channels_volt,df_tmp],axis=1)


    PMT_error=False
    PMT_error_str = ""
    PMT_error_handling_str = ""
    most_common_V = max(set(ch_volt), key=ch_volt.count)

    # check if all channels are at set-voltage
    for i in np.arange(0,np.size(channels,0)):
        if ch_volt[i]!=most_common_V: # if channel voltage is not set voltage
            PMT_error =True
            PMT_error_str = PMT_error_str + channels.loc[3+i,'$PnN'] + ', '

    if PMT_error:
        PMT_error_handling_str = 'Channel(s) will be ignored'
    else:
        PMT_error_str = 'OK'

    # check if the laser delay is set to zero for all channels
    laser_delay_error = df_laser[df_laser['laser delay']!=0]
    if laser_delay_error.empty:
        Laser_delay_error_str = 'OK'
    else:
        Laser_delay_error_str = 'Error: '
        for tmp in laser_delay_error['laser name']:
            Laser_delay_error_str = Laser_delay_error_str + tmp + ', '

    # build DataFrame
    if first_data_set:
        try:
            Device=meta['$CYT']
        except:
            Device='None'

        try:
            DeviceID=meta['CYTNUM']
        except:
            DeviceID='None'

        df_meta = pd.DataFrame({'File':[file_name],'Device':[Device],'Device ID':[DeviceID],'Date':[meta['$DATE']],'Set voltage':[str(most_common_V)],'PMT voltage':[PMT_error_str],'Error handling':[PMT_error_handling_str],'Laser delay':Laser_delay_error_str})
    else:
        df_tmp = pd.DataFrame({'File':[file_name],'Device':[Device],'Device ID':[DeviceID],'Date':[meta['$DATE']],'Set voltage':[str(most_common_V)],'PMT voltage':[PMT_error_str],'Error handling':[PMT_error_handling_str],'Laser delay':Laser_delay_error_str})
        df_meta = pd.concat([df_meta,df_tmp],axis=0,sort=False)
    first_data_set = False
    ctr += 1
    col_names = df_data.columns
    ind = np.arange(df_meta.shape[0])
    df_meta.index = ind

    df_tmp = df_meta[df_meta['Device ID'] != df_meta['Device ID'][0]]
    if df_tmp.empty:
        inconsistent_device = False
        dev_str = df_meta['Device'][0]
        dev_ID_str = df_meta['Device ID'][0]
    else:
        inconsistent_device = True
        dev_str = 'data from different flow cytometers'
        dev_ID_str = 'data from different flow cytometers'
    df_tmp = df_meta[df_meta['Date'] != df_meta['Date'][0]]
    if df_tmp.empty:
        inconsistent_date = False
        date_str = df_meta['Date'][0]
    else:
        inconsistent_date = True
        date_str = 'data from different dates'
    gui.progress("{:f}".format(counter/NN*100))
    counter += 1

df_tmp = pd.DataFrame({'lib_files':lib_files},index=ch_volt_from_name)
df_file_count = pd.concat([df_file_count,df_tmp],axis=1)

df_volt_compare = pd.concat([df_volt_compare,df_file_count],axis=1)
df_volt_compare = df_volt_compare.fillna(False)

df_total_ramp = pd.DataFrame(index=ramp_files, columns=['Ramp total'])
df_total_lib = pd.DataFrame(index=lib_files, columns=['Lib total'])
df_total_ramp['Ramp total']=ramp_total
df_total_lib['Lib total']=lib_total


# ----------------------
# Export consistency-check to Excel
# ----------------------
def init_excel(fname, sheets, titels):
    wb = Workbook()

    for nr, name in enumerate(sheets):
        wb.create_sheet(name)
        ws = wb[name]
        if titels[nr]!='':
            ws['A1'] = titels[nr]

            # Title banner from A1:Z1
            cell_range = ['%s1' % get_column_letter(nr) for nr in range(1,27)]
            # Banner style
            blue_fill = styles.PatternFill("solid", fgColor="4398ED")
            White_color_font = 'FFFFFF'
            White_font = styles.Font(size=28, bold=True, color=White_color_font)

            for i in cell_range:
                cell = ws[i]

                cell.fill = blue_fill
                cell.font = White_font

    wb.remove(wb['Sheet'])
    wb.save(fname)

# Style
thin = styles.Side(border_style="thin", color="000000")
border1 = styles.Border(top=thin, left=thin, right=thin, bottom=thin)
font = styles.Font(b=True, color="000000")
fill = styles.PatternFill("solid", fgColor="FFFFFF")

# Condition styles
red_color_font = '9c0103'
red_font = styles.Font(size=12, bold=True, color=red_color_font)
green_color_font = '228B22'
green_font = styles.Font(size=12, bold=False, color=green_color_font)

# Check for Evaluation folder, if not -> create
directory = os.path.join(data_folder,'Evaluation')
if not os.path.exists(directory):
    os.makedirs(directory)

print('ConsistencyCheck.xlsx')
fname = os.path.join(directory,'ConsistencyCheck.xlsx')
if os.path.isfile(fname):
    #ToDo: GUI / Error 
    # print('ConsistencyCheck.xlsx is allready present in directory. \n Overwrite (y/n)')
    if gui.confirm("ConsistencyCheck.xlsx is allready present in directory. \n Overwrite?") != 'OK':
        sys.exit()

# ----------------------
# Initialization of excel
# ----------------------
sheets = ['Overview', 'Channels', 'Sample', 'Error log']
titels = ['Consistency overview', 'Channels: Voltages', 'Samples: intensities', '']
# sheets = ['Overview', 'Channels', 'Sample', 'Error log1', 'Error log2']
# titels = ['Consistency overview', 'Channels: Voltages', 'Samples: intensities', '', '']
init_excel(fname, sheets, titels)


# ----------------------
# Load Writer and sheets
# ----------------------
wb = load_workbook(fname)
writer = pd.ExcelWriter(fname, engine='openpyxl')
writer.book = wb
writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)


# ----------------------
# Write overview page
# ----------------------
ws = writer.sheets[sheets[0]]
Line1=13

# === Laser info ===
df_laser = df_laser[['laser name','laser delay','area scaling factor']]
df_laser.to_excel(writer, sheet_name=sheets[0], startrow=Line1)

# Conditional formatting
for row in np.arange(Line1+2, Line1+2+df_laser.shape[0]):
    cellname='C'+str(row)
    Treshold=0
    ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='notEqual',formula=[Treshold], font=red_font))
    ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='equal',formula=[Treshold], font=green_font))

# title
laser_table_pos='A'+str(Line1+1)
ws[laser_table_pos]='Lasers'
style_cell(ws, laser_table_pos, border=border1, fill=fill, font=font)
cellrange= 'B' + str(Line1+2) + ':' + get_column_letter(df_laser.shape[1]+1) +str(Line1+1+df_laser.shape[0])
style_range(ws, cellrange, border=border1)


# Column width
ws.column_dimensions['A'].width = 20
ws.column_dimensions['B'].width = 16
ws.column_dimensions['C'].width = 16
ws.column_dimensions['D'].width = 20
ws.column_dimensions['E'].width = 20
ws.column_dimensions['G'].width = 20
ws.column_dimensions['H'].width = 20

# === Cytometer info ===
style_range(ws, 'A3:B3', border=border1, fill=fill, font=font)
style_range(ws, 'A4:B6', border=border1)
ws['A3'] = 'Cytometer'
ws['A4'] = 'Device'
ws['A5'] = 'Device ID'
ws['A6'] = 'Date'
ws['B4'] = dev_str
ws['B5'] = dev_ID_str
ws['B6'] = date_str

# QuantiFlash settings
style_range(ws, 'D3:E3', border=border1, fill=fill, font=font)
style_range(ws, 'D4:E9', border=border1)
ws['D3'] = 'QuantiFlash settings'
ws['D4'] = 'Start intensity'
ws['D5'] = 'Stop intensity'
ws['D6'] = 'Step size'
ws['D7'] = 'Trigger intensity'
ws['D8'] = 'Event rate'
ws['D9'] = 'Nr. of samples'
ws['E4'] = startInt
ws['E5'] = stopInt
ws['E6'] = stepInt
ws['E7'] = triggerInt
ws['E8'] = countRate
ws['E9'] = nSamples

# Files found
style_range(ws, 'A8:B8', border=border1, fill=fill, font=font)
style_range(ws, 'A9:B11', border=border1)
ws['A9'] = 'Files'
ws['A10'] = 'Ramp'
ws['A11'] = 'Lib'
ws['A12'] = 'Directory:'
ws['B10'] = np.size(ramp_files)
ws['B11'] = np.size(lib_files)
ws['B12'] = data_folder

Line2 = Line1 + df_laser.shape[0] + 3

# Channel info
Channels_table_pos='A'+str(Line2+1)
ws[Channels_table_pos]='Channels'
style_cell(ws, Channels_table_pos, border=border1, fill=fill, font=font)
df_channels.to_excel(writer, sheet_name=sheets[0], startrow=Line2)
cellrange= 'B' + str(Line2+2) + ':' + get_column_letter(df_channels.shape[1]+1)+str(Line2+1+df_channels.shape[0])
style_range(ws, cellrange, border=border1)

Line3 = Line2 + df_channels.shape[0] + 3

# display all files:
df_tmp = df_meta.drop(['Date','Device','Device ID'],axis=1)

df_tmp = df_tmp.set_index('File')
df_tmp = df_tmp[['Set voltage','Laser delay','PMT voltage','Error handling']]
if ~df_tmp.empty:
    df_tmp.to_excel(writer, sheet_name=sheets[0], startrow=Line3)
cellrange= 'B' + str(Line3+2) + ':' + get_column_letter(df_tmp.shape[1]+1) +str(Line3+1+df_tmp.shape[0])
style_range(ws, cellrange, border=border1)


# ----------------------
# Write channels sheet
# ----------------------
Line1 = 3
df_channels_volt.T.to_excel(writer, sheet_name=sheets[1], startrow=Line1)
ws = writer.sheets[sheets[1]]

# Conditional formatting
for row in np.arange(Line1+2, Line1+2+df_channels_volt.T.shape[0]):
    for column in np.arange(2, 2+df_channels_volt.T.shape[1]):

        cellname=get_column_letter(column)+str(row)
        most_common=df_channels_volt.iloc[:,row-Line1-2].mode()[0]
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='notEqual',formula=[most_common], font=red_font))
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='equal',formula=[most_common], font=green_font))

cellrange= 'B' + str(Line1+2) + ':' + get_column_letter(df_channels_volt.T.shape[1]+1) +str(Line1+1+df_channels_volt.T.shape[0])
style_range(ws, cellrange, border=border1)
Line2=Line1+df_channels_volt.T.shape[0]+3


#1 subtitle: Files
position = 'A'+str(Line1+1)
ws[position] = 'Files'
style_cell(ws, position, border=border1, fill=fill, font=font)

#2 subtitle
position = 'B'+str(Line1)
cellrange = 'B'+str(Line1)+':C'+str(Line1)
ws[position] = 'Channels'
style_range(ws, cellrange, border=border1, fill=fill, font=font)

#3 Info: Red/Green
position = 'A'+str(Line2)
cellrange = 'A'+str(Line2)+':K'+str(Line2)
ws[position] = 'Info: Red values are voltages that do not match the most common voltage of that file and are most likely wrong. '
style_range(ws, cellrange, border=border1, fill=fill, font=font)


# ----------------------
# Write excel Sample sheet
# ----------------------
df_Int_orig = df_Int
df_Int.columns=ramp_files
#df_Int.index = Int_step[:-1]
df_Int.index = Int_step
df_Int=df_Int.T
df_Int_Off=df_Int.Off
df_Int=df_Int.drop(labels='Off',axis=1)


Line1=3
voltage_list=[temp2[3:-4] for temp2 in lib_files]
df_total_laser = pd.DataFrame([list(df_total_lib['Lib total']),list(df_Int_Off)]).T
df_total_laser.index=voltage_list
df_total_laser.columns=['On','Off']
df_total_ramp.columns=['Samples']
df_total_lib.columns=['Samples']

df_Int.to_excel(writer, sheet_name=sheets[2], startrow=Line1)
ws = writer.sheets[sheets[2]]

# Conditional formatting
for row in np.arange(Line1+2, Line1+2+df_Int.shape[0]):
    for column in np.arange(2,2+df_Int.shape[1]):

        cellname=get_column_letter(column)+str(row)
        Treshold=0.95*nSamples
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='lessThanOrEqual',formula=[Treshold], font=red_font))
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='greaterThan',formula=[Treshold], font=green_font))


cellrange= 'B' + str(Line1+2) + ':' + get_column_letter(df_Int.shape[1]+1) +str(Line1+1+df_Int.shape[0])
style_range(ws, cellrange, border=border1)


Line2=Line1+df_Int.shape[0]+3


df_total_ramp.to_excel(writer, sheet_name=sheets[2], startrow=Line2, startcol=0)
cellrange= 'B' + str(Line2+2) + ':' + get_column_letter(1+df_total_ramp.shape[1]) +str(Line2+1+df_total_ramp.shape[0])
style_range(ws, cellrange, border=border1)

# Conditional formatting
for row in np.arange(Line2+2, Line2+2+df_total_ramp.shape[0]):
    cellname='B'+str(row)
    Treshold=0.95*nSamples*np.shape(Int_step)[0]
    ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='lessThanOrEqual',formula=[Treshold], font=red_font))
    ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='greaterThan',formula=[Treshold], font=green_font))


df_total_lib.to_excel(writer, sheet_name=sheets[2], startrow=Line2, startcol=3)
cellrange= 'D' + str(Line2+2) + ':' + get_column_letter(4+df_total_lib.shape[1]) +str(Line2+1+df_total_ramp.shape[0])
style_range(ws, cellrange, border=border1)

# Conditional formatting
for row in np.arange(Line2+2, Line2+2+df_total_lib.shape[0]):
        cellname='E'+str(row)
        Treshold=0.95*nSamples
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='lessThanOrEqual',formula=[Treshold], font=red_font))
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='greaterThan',formula=[Treshold], font=green_font))


df_total_laser.to_excel(writer, sheet_name=sheets[2], startrow=Line2, startcol=6)
cellrange= 'H' + str(Line2+2) + ':' + get_column_letter(7+df_total_laser.shape[1]) +str(Line2+1+df_total_ramp.shape[0])
style_range(ws, cellrange, border=border1)

# Conditional formatting
for row in np.arange(Line2+2, Line2+2+df_total_laser.shape[0]):
    for column in np.arange(8,8+df_total_laser.shape[1]):

        cellname=get_column_letter(column)+str(row)
        Treshold=0.95*nSamples
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='lessThanOrEqual',formula=[Treshold], font=red_font))
        ws.conditional_formatting.add(cellname, formatting.rule.CellIsRule(operator='greaterThan',formula=[Treshold], font=green_font))

# Create subtitles
#1 subtitle: Files
position = 'A'+str(Line1+1)
ws[position] = 'Files'
style_cell(ws, position, border=border1, fill=fill, font=font)

#2 subtitle: Intensities
position = 'B'+str(Line1)
cellrange = 'B'+str(Line1)+':C'+str(Line1)
ws[position] = 'Intensities [dB]'
style_range(ws, cellrange, border=border1, fill=fill, font=font)

#3 subtitle: Ramp measurement
position = 'A'+str(Line2)
cellrange = 'A'+str(Line2)+':B'+str(Line2)
ws[position] = 'Ramp measurement:'
style_range(ws, cellrange, border=border1, fill=fill, font=font)

#4 subtitle: Files
position = 'A'+str(Line2+1)
ws[position] = 'Files'
style_cell(ws, position, border=border1, fill=fill, font=font)

#5 subtitle: Lib measurement
position = 'D'+str(Line2)
cellrange = 'D'+str(Line2)+':E'+str(Line2)
ws[position] = 'LIB measurement:'
style_range(ws, cellrange, border=border1, fill=fill, font=font)

#6 subtitle: Files
position = 'D'+str(Line2+1)
ws[position] = 'Files'
style_cell(ws, position, border=border1, fill=fill, font=font)

#7 subtitle: Voltages
position = 'G'+str(Line2+1)
ws[position] = 'Voltage'
style_cell(ws, position, border=border1, fill=fill, font=font)

#8 subtitle: Number of samples Laser on/off
position = 'G'+str(Line2)
cellrange = 'G'+str(Line2)+':I'+str(Line2)
ws[position] = 'Nr. of samples laser on/off:'
style_range(ws, cellrange, border=border1, fill=fill, font=font)


# ----------------------
# Write Error log1 sheet
# ----------------------
# Line1 = 3

df_volt_compare = df_volt_compare*1
df_volt_compare.to_excel(writer, sheet_name=sheets[3])
ws = writer.sheets[sheets[3]]

# ----------------------
# Write Error log2 sheet
# ----------------------
# Line1 = 3

# df_complete = df_Int_orig.isna().sum()==0
# df_complete = pd.DataFrame(df_complete)
# df_complete.rename({'0':'complete'},inplace=True)
# df_complete = df_complete*1
# df_complete
# df_volt_compare.to_excel(writer, sheet_name=sheets[4])
# ws = writer.sheets[sheets[4]]


writer.save()
wb.close()

# ==========================================
# Part 1 of evaluation done
# ------------------------------------------
print('------------------------------------')
print('Part 1 of evaluation done.')
print('')
print('Check Consisteny excel file')
print('------------------------------------')
print('Run Evaluation_SNR_DNR.py to continue ')
